"""The central logger module of the application"""
import sys
from loguru import logger

# Use the loguru as logger
app_logger = logger


def get_level_choices():
    """Provides the list of valid log level names"""
    return ["critical", "error", "warning", "success", "info", "debug", "trace"]


def get_format_choices():
    """Provides the list of valid log format names"""
    return ["text", "json"]


def get_logger():
    """Returns with the application logger"""
    return app_logger


def init_logger(log_level: str, log_format: str):
    """Configures the logger instance"""
    # Remove the default logger before configuring
    app_logger.remove()

    # Create a new sink instance, and set the format and level
    if log_format.upper() == "JSON":
        app_logger.add(sys.stderr, level=log_level.upper(), serialize=True)
    else:
        app_logger.add(sys.stderr, level=log_level.upper())

    return app_logger



"""The application base class"""
from abc import ABC, abstractmethod
import asyncio
from typing import Dict, Optional, Any
from ..logger.logger import init_logger
from .signals import DelayedKeyboardInterrupt


class ApplicationBase(ABC):
    """
    The ApplicationBase class
    """

    def __init__(self, config):
        """
        Initializes the new application object
        """
        self._loop = None  # type: Optional[asyncio.AbstractEventLoop]
        self._wait_event = None  # type: Optional[asyncio.Event]
        self._wait_task = None  # type: Optional[asyncio.Task]
        self.logger = init_logger(config.LOG_LEVEL, config.LOG_FORMAT)
        self.config = config

    def run(self):
        """
        Run the application
        """
        self._loop = asyncio.new_event_loop()

        try:
            # Shield start() from termination.
            try:
                with DelayedKeyboardInterrupt(self.logger):
                    self._start()
            # If there was an attempt to terminate the application,
            # the KeyboardInterrupt is raised AFTER the _start() finishes its job.
            #
            # In that case, the KeyboardInterrupt is re-raised and caught in
            # exception handler below and _stop() is called to clean all resources.
            #
            # Note that it might be generally unsafe to call stop() methods
            # on objects that are not started properly.
            # This is the main reason why the whole execution of _start()
            # is shielded.

            except KeyboardInterrupt:
                self.logger.info("Application.run: got KeyboardInterrupt during start")
                raise

            # Application is started now and is running.
            # Wait for a termination event infinitelly.

            self.logger.info("Application.run: entering wait loop")
            self._wait()
            self.logger.info("Application.run: exiting wait loop")

        except KeyboardInterrupt:
            # The stop() is also shielded from termination.
            try:
                with DelayedKeyboardInterrupt(self.logger):
                    self._stop()
            except KeyboardInterrupt:
                self.logger.info("Application.run: got KeyboardInterrupt during stop")

    @abstractmethod
    async def start(self):
        """Starts the application"""

    @abstractmethod
    async def stop(self):
        """Shuts down the application"""

    async def wait(self):
        """
        Wait until the application got stop signal
        """
        self._wait_event = asyncio.Event()
        self._wait_task = asyncio.create_task(self._wait_event.wait())
        await self._wait_task

    def _start(self):
        self._loop.run_until_complete(self.start())

    def _stop(self):
        self._loop.run_until_complete(self.stop())

        # Because we want clean exit, we patiently wait for completion
        # of the _wait_task (otherwise this task might get cancelled
        # in the _cancel_all_tasks() method - which wouldn't be a problem,
        # but it would be dirty).
        #
        # The _wait_event & _wait_task might not exist if the application
        # has been terminated before calling _wait(), therefore we have to
        # carefully check for their presence.

        if self._wait_event:
            self._wait_event.set()

        if self._wait_task:
            self._loop.run_until_complete(self._wait_task)

        #
        # Before the loop is finalized, we setup an exception handler that
        # suppresses several nasty exceptions.
        #
        # ConnectionResetError
        # --------------------
        # This exception is sometimes raised on Windows, possibly because of a bug in Python.
        #
        # ref: https://bugs.python.org/issue39010
        #
        # When this exception is raised, the context looks like this:
        # context = {
        #     'message': 'Error on reading from the event loop self pipe',
        #     'exception': ConnectionResetError(
        #         22, 'The I/O operation has been aborted because of either a thread exit or an application request',
        #         None, 995, None
        #       ),
        #     'loop': <ProactorEventLoop running=True closed=False debug=False>
        # }
        #
        # OSError
        # -------
        # This exception is sometimes raised on Windows - usually when application is
        # interrupted early after start.
        #
        # When this exception is raised, the context looks like this:
        # context = {
        #     'message': 'Cancelling an overlapped future failed',
        #     'exception': OSError(9, 'The handle is invalid', None, 6, None),
        #     'future': <_OverlappedFuture pending overlapped=<pending, 0x1d8937601f0>
        #                 cb=[BaseProactorEventLoop._loop_self_reading()]>,
        # }
        #

        def __loop_exception_handler(_, context: Dict[str, Any]):
            if isinstance(context["exception"], ConnectionResetError):
                self.logger.info(
                    "Application._stop.__loop_exception_handler: suppressing ConnectionResetError"
                )
            elif isinstance(context["exception"], OSError):
                self.logger.info(
                    "Application._stop.__loop_exception_handler: suppressing OSError"
                )
            else:
                self.logger.info(
                    f"Application._stop.__loop_exception_handler: unhandled exception: {context}"
                )

        self._loop.set_exception_handler(__loop_exception_handler)

        try:
            # Cancel all remaining uncompleted tasks.
            # We should strive to not make any, but mistakes happen and laziness
            # is also a thing.
            #
            # Generally speaking, cancelling tasks shouldn't do any harm (unless
            # they do...).
            self._cancel_all_tasks()

            # Shutdown all active asynchronous generators.
            self._loop.run_until_complete(self._loop.shutdown_asyncgens())
        finally:
            # ... and close the loop.
            self.logger.info("Application._stop: closing event loop")
            self._loop.close()

    def _wait(self):
        self._loop.run_until_complete(self.wait())

    def _cancel_all_tasks(self):
        """
        Cancel all tasks in the loop.

        This method injects an asyncio.CancelledError exception
        into all tasks and lets them handle it.

        Note that after cancellation, the event loop is executed again and
        waits for all tasks to complete the cancellation.  This means that
        if some task contains code similar to this:

        >>> except asyncio.CancelledError:
        >>>     await asyncio.Event().wait()

        ... then the loop doesn't ever finish.
        """

        #
        # Code kindly borrowed from asyncio.run().
        #

        to_cancel = asyncio.tasks.all_tasks(self._loop)
        self.logger.info(
            f"Application._cancel_all_tasks: cancelling {len(to_cancel)} tasks ..."
        )

        if not to_cancel:
            return

        for task in to_cancel:
            task.cancel()

        self._loop.run_until_complete(
            asyncio.tasks.gather(*to_cancel, loop=self._loop, return_exceptions=True)
        )

        for task in to_cancel:
            if task.cancelled():
                continue

            if task.exception() is not None:
                self._loop.call_exception_handler(
                    {
                        "message": "unhandled exception during Application.run() shutdown",
                        "exception": task.exception(),
                        "task": task,
                    }
                )



from setuptools import find_packages, setup
import pathlib
import os

# Package metadata
# ----------------
APP_NAME = "py-12f-common"
APP_DESCRIPTION = """This repository holds those infrastructure-level modules, that every application requires that follows the core 12-factor principles."""

# Get the long description from the README file
HERE = pathlib.Path(__file__).parent.resolve()
LONG_DESCRIPTION = (HERE / "README.md").read_text(encoding="utf-8")

URL = "https://github.com/tombenke/py-12f-common"
EMAIL = "tombenke@gmail.com"
AUTHOR = "Tamás Benke"
LICENSE = "MIT"
REQUIRES_PYTHON = ">=3.8"

# What packages are required for this module to be executed?
REQUIRED = [
    "argparse",
    "dataclasses",
    "loguru",
]

DEV_REQUIREMENTS = [
    "build",
    "coverage",
    "coverage-badge",
    "black",
    "pdoc",
    "pydeps",
    "pylint",
]

setup(
    name=APP_NAME,
    version=os.getenv("VERSION", "1.0.0"),
    description=APP_DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    license=LICENSE,
    packages=find_packages(exclude=("tests", "docs")),
    include_package_data=True,
    install_requires=REQUIRED,
    extras_require={"dev": DEV_REQUIREMENTS},
    entry_points={},
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
)



"""Generic Config related classes"""
import argparse
from dataclasses import dataclass, astuple
import os
import re


@dataclass
class CliEntry:
    """
    Details of a ConfigEntry about how to setup the CLI arg parser
    NOTE: the properties hold by this class are strongly related to
    """

    short_flag: str
    name: str
    entry_type: type = str
    action: str = "store"
    choices: list = None

    def __iter__(self):
        """
        Return with an iterable collection of the object properties
        """
        return iter(astuple(self))


@dataclass
class ConfigEntry:
    """
    Config parameter deswcriptor class
    """

    name: str
    help_text: str = ""
    default: str = ""
    cli: CliEntry = None

    def __iter__(self):
        """
        Return with an iterable collection of the object properties
        """
        return iter(astuple(self))


@dataclass
class Config:
    """
    Configuration class that represent the actual config parameter set of the application.
    The class provides method for resolving the config parameters
    from env variables as well as from CLI parameters.
    """

    def __init__(self, app_name, app_description, config_entries):
        """
        Constructor for the application config
        It fills the object with the properties given by the `config_entries` argument.
        It also sets the initial value of each parameter from the environment, if it is found
        or set to its default value if the corresponding environment var is missing.
        """
        self.app_name = app_name
        self.app_description = app_description
        self.config_entries = config_entries

        for config_entry in config_entries:
            (name, _, default, _) = config_entry
            self.__dict__[name] = os.environ.get(name, default)

    def apply_parameters(self, parameters):
        """
        Overwrites the actual values of the configuration properties with the values given by the
        `parameters` dictionary. Typically usage is, when the CLI argument parser wants to overwrite
        the default, or environment values.
        """
        for k in parameters:
            self.__dict__[k] = parameters[k]

    def apply_cli_args(self, argv):
        """
        Take the actual CLI parameters according to the definitions in `self.config_entries`
        parse them, and apply them to the actual properties of the config object.
        """
        args = self.get_the_cli_args(argv)
        self.apply_parameters(args)

    def get_the_cli_args(self, argv):
        """
        Parse the CLI parameters, and returns with them as a dictionary
        """
        parser = argparse.ArgumentParser(
            prog=self.app_name, description=self.app_description
        )

        # Build the CLI args parser
        for config_entry in self.config_entries:
            (name, help_text, _, cli) = config_entry
            default = self.__dict__[name]
            if cli is not None:
                (short_flag, name, entry_type, action, choices) = cli
                if action in ["store_false", "store_true"]:
                    # It is a boolean flag
                    parser.add_argument(
                        short_flag, name, help=help_text, default=default, action=action
                    )
                else:
                    if choices is not None:
                        # It is a selection from the list of choices
                        parser.add_argument(
                            short_flag,
                            name,
                            type=entry_type,
                            help=help_text,
                            default=default,
                            action=action,
                            choices=choices,
                        )
                    else:
                        # It is a normal flag or positional parameter
                        parser.add_argument(
                            short_flag,
                            name,
                            type=entry_type,
                            help=help_text,
                            default=default,
                            action=action,
                        )

        # Parse the CLI args
        args = vars(parser.parse_args(argv))

        # Build a dictionary of actual CLI arg values, using the original parameter name
        results = {}
        for config_entry in self.config_entries:
            # Take the original entry name
            (entry_name, _, _, cli) = config_entry

            if cli is not None:
                # Take the parameter value of the CLI arg
                (_, name_or_flag, _, _, _) = cli
                name = re.sub(r"^--", "", name_or_flag).replace("-", "_")

                # Add the value to the results with the original name
                results[entry_name] = args[name]

        return results

    def dump(self):
        """
        Prints the actual values of the config parameters to the console
        """
        print("\nConfig:")
        for config_entry in self.config_entries:
            (name, _, _, _) = config_entry
            print(f"  {name}: '{self.__dict__[name]}'")

    def __str__(self):
        """
        Returns with the string representation of the object
        """
        str_repr = ""
        for key, value in self.__dict__.items():
            str_repr += f"{key}='{value}', "

        return str_repr



"""
DelayedKeyboardInterrupt implementation.
"""
import os
import signal


__all__ = [
    "SIGNAL_TRANSLATION_MAP",
]

SIGNAL_TRANSLATION_MAP = {
    signal.SIGINT: "SIGINT",
    signal.SIGTERM: "SIGTERM",
}


class DelayedKeyboardInterrupt:
    """
    A context provider class, that makes possible to postpone the handling of exceptions
    that may occur inside the context.
    """

    def __init__(self, logger, propagate_to_forked_processes=None):
        """
        Constructs a context manager that suppresses SIGINT & SIGTERM signal handlers
        for a block of code.
        The signal handlers are called on exit from the block.
        Inspired by: https://stackoverflow.com/a/21919644
        :param propagate_to_forked_processes: This parameter controls behavior of this context manager
        in forked processes.
        If True, this context manager behaves the same way in forked processes as in parent process.
        If False, signals received in forked processes are handled by the original signal handler.
        If None, signals received in forked processes are ignored (default).
        """
        self.logger = logger
        self._pid = os.getpid()
        self._propagate_to_forked_processes = propagate_to_forked_processes
        self._sig = None
        self._frame = None
        self._old_signal_handler_map = None

    def __enter__(self):
        self._old_signal_handler_map = {
            sig: signal.signal(sig, self._handler)
            for sig, _ in SIGNAL_TRANSLATION_MAP.items()
        }

    def __exit__(self, exc_type, exc_val, exc_tb):
        for sig, handler in self._old_signal_handler_map.items():
            signal.signal(sig, handler)

        if self._sig is None:
            return

        self._old_signal_handler_map[self._sig](self._sig, self._frame)

    def _handler(self, sig, frame):
        self._sig = sig
        self._frame = frame

        #
        # Protection against fork.
        #
        if os.getpid() != self._pid:
            if self._propagate_to_forked_processes is False:
                self.logger.error(
                    f"!!! DelayedKeyboardInterrupt._handler: {SIGNAL_TRANSLATION_MAP[sig]} received; "
                    f"PID mismatch: {os.getpid()=}, {self._pid=}, calling original handler"
                )
                self._old_signal_handler_map[self._sig](self._sig, self._frame)
            elif self._propagate_to_forked_processes is None:
                self.logger.error(
                    f"!!! DelayedKeyboardInterrupt._handler: {SIGNAL_TRANSLATION_MAP[sig]} received; "
                    f"PID mismatch: {os.getpid()=}, ignoring the signal"
                )
                return
            # elif self._propagate_to_forked_processes is True:
            #   ... passthrough

        self.logger.error(
            f"!!! DelayedKeyboardInterrupt._handler: {SIGNAL_TRANSLATION_MAP[sig]} received; delaying KeyboardInterrupt"
        )
